/*

  collision.c

  Functions for collision detection.

    Copyright (C) 1993 Rob Shillingsburg
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


*/

#include "polyroids.h"


/*** File-scope variable definitions: ****************************************/

/* Holds segment structures for the ship's segments. */
Segment shipSegs[3];

/*** Internal function declarations: ****************************************/

int CheckShipCollisions(Polyroid *polyroid, Segment *currSegPtr);


/*** Interface function definitions: *****************************************/

void CollisionDetect(float deltaT)
{
  int i, shipHit = FALSE;
  Segment laserSeg;

  /* Store a segment for the laser. */
  laserSeg.pt1 = ship.vertices[0];
  laserSeg.pt2 = laserEnd;
  laserSeg.slope = CalculateSlope(laserSeg.pt1, laserSeg.pt2);

  /* Store 3 segements for the ship. */
  for (i = 0; i < 3; i++) {
    shipSegs[i].pt1 = ship.vertices[0];
    shipSegs[i].pt2 = ship.vertices[(i+1) % 3];
    shipSegs[i].slope = CalculateSlope(shipSegs[i].pt1, shipSegs[i].pt2);
  }

  for (i = 0; i < numPolyroids; i++) {
    Polyroid *polyroid = polyroids[i];
    Segment currSeg;
    int side;
    int laserHit = FALSE;
    
    if (polyroid->dead) continue;  /* Dead ones don't collide. */

    for (side = 0; side < polyroid->numSides; side++) {
      /* Set up a segment for this side of the polyroid. */
      currSeg.pt1 = polyroid->vertices[side];
      currSeg.pt2 = polyroid->vertices[(side + 1) % polyroid->numSides];
      currSeg.slope = CalculateSlope(currSeg.pt1, currSeg.pt2);

      /* See if the side hits the laser or the ship. */
      laserHit = laserHit ||
	(currentLaser != -1 && SegmentsIntersect(&currSeg, &laserSeg)); 
      shipHit = shipHit || CheckShipCollisions(polyroid, &currSeg);
    }

    /* Note whether the polyroid is being hit, so we can draw effects. */
    polyroid->beingHit = laserHit;

    /* Also check a segment through the center of the polyroid,
       so the ship can't `hide' inside a polyroid. */
    currSeg.pt1 = polyroid->vertices[0];
    currSeg.pt2 = polyroid->vertices[polyroid->numSides/2];
    currSeg.slope = CalculateSlope(currSeg.pt1, currSeg.pt2);
    shipHit = shipHit || CheckShipCollisions(polyroid, &currSeg);

    if (laserHit) {
      /* Push the polyroid (firehose effect). */
      polyroid->dx += deltaT * ship.shotThrust * cos(ship.angle) *
	((polyroid->kind == currentLaser) ? LASER_MATCH_FACTOR : 1);
      polyroid->dy += deltaT * ship.shotThrust * sin(ship.angle) *
	((polyroid->kind == currentLaser) ? LASER_MATCH_FACTOR : 1);

      /* Decrease the polyroid's spin rate. */
      polyroid->dTheta += deltaT *
	((polyroid->kind == currentLaser) ? LASER_MATCH_FACTOR : 1) *
	  ((polyroid->ddTheta > 0) ? -ship.shotStrength : ship.shotStrength);

      if (polyroid->dTheta * polyroid->ddTheta < 0) {
	/* Polyroid loses a side.  Too bad. */
	polyroid->numSides--;
	polyroid->dTheta = MAX_POLYROID_SPIN - ship.shotStrength * deltaT;
	polyroid->ddTheta = fabs(polyroid->ddTheta);
	
	if (polyroid->numSides < 3) {
	  /* Polyroid is dead.  Awwww... */
	  KillPolyroid(polyroid);
	}
      }
    }
  }

  if (shipHit) {
    ship.shields -= deltaT * COLLISION_PENALTY;
    dead = TRUE;  /* Not really dead, just being hit. */
    if (ship.shields < 0.0) {  /* Game over. */
      ship.shields = 0;  /* Don't want to display negative shields. */
      EndGame();
    }
  }
}


/*** Internal function definitions: ******************************************/


/* Returns TRUE if ship is is contact with the given segment. */
int CheckShipCollisions(Polyroid *polyroid, Segment *currSegPtr)
{
  int shipSide;

  for (shipSide = 0; shipSide < 3; shipSide++) {
    if (SegmentsIntersect(currSegPtr, &shipSegs[shipSide])) {
      return TRUE;
    }
  }
  
  return FALSE;
}

