/*

  polyroids.h

  Header file for the polyroids program.

    Copyright (C) 1993 Rob Shillingsburg
    Copyright (C) 2002 Luka Frelih <luka@mail.ljudmila.org>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


*/

#ifdef __APPLE__
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#include <GLUT/glut.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#endif 

#include "port.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

// include <GL/device.h>
// include <fmclient.h>

/*** Constants: **************************************************************/

#define PI                     3.1415927
#define ROOT_2                 1.4142136

#define X 0                 /* Indices for point arrays. */
#define Y 1

#define TRUE 1
#define FALSE 0

#define J_LASER 0            /* Indices for laser tables. */
#define K_LASER 1
#define L_LASER 2
#define SEMI_LASER 3

/* Some packed colors. */
#define RGB_BLACK 		0, 0, 0
#define RGB_RED			1, 0, 0
#define RGB_GREEN               0, 1, 0
#define RGB_BLUE                0, 0, 1
#define RGB_CYAN		0, 1, 1
#define RGB_YELLOW		1, 1, 0
#define RGB_MAGENTA             1, 0, 1
#define RGB_WHITE		1, 1, 1

/* Window position and dimensions. */
#define XSIZE 1024
#define YSIZE 768
#define XPOS 0
#define YPOS 0

#define MAX_POLYROIDS 20         /* Max number of polyroids on the screen. */
#define POLYROID_RADIUS 35  /* Distance from center to vertex of a polyroid. */
#define POLYROID_MIN_RADIUS 5    /* Min radius for a dead polyroid. */
#define SHRINK_SPEED 8           /* In pixels per second. */
#define MAX_POLYROID_SIDES 6     /* Max number of sides of a polyroid. */
#define MAX_POLYROID_SPIN 4*PI   /* Spin rate when polyroid pops a new side. */
#define PER_SIDE_POLYROID_VELOCITY 50  /* Velocity of a 1-sided polyroid. */
#define PER_SIDE_POLYROID_ACCELERATION 100 /* Accel of a 1-sided polyroid. */



#define SHIP_RADIUS 15            /* Size of the ship. */
#define SHIP_CORNER_ANGLE 4*PI/5  /* Between 0 and PI */
#define SHIP_COLOR RGB_WHITE      /* Ship's color. */
#define SHIP_THRUST 500           /* Pixels per second per second. */
#define DRAG_COEFF .4             /* Fraction of velocity lost per second. */
#define SHIP_TURN 3*PI/2          /* Radians per second. */
#define MAX_SHIP_VELOCITY 1000    /* Manhattan... */
#define BASE_SHOT_STRENGTH 5*PI   /* Radians per second. */
#define INC_SHOT_STRENGTH .25*PI   /* Radians per second. */
#define BASE_SHOT_THRUST 500      /* Pixels per second per second. */
#define INC_SHOT_THRUST 25        /* Pixels per second per second. */
#define LASER_MATCH_FACTOR 2      /* Bonus to strength and thrust for
				     hitting with the proper laser. */
#define INITIAL_CONCUSSIONS 3     /* Starting number of concussions. */
#define CONCUSSION_STEPS 15       /* Number of waves in a concussion. */
#define CONCUSSION_RADIUS 15      /* Distance between successive waves. */
#define EXPLOSION_STRENGTH 100000 /* Pixels per second per second. */
#define BASE_REGEN_RATE 0.006     /* Fraction of shield regained per second. */
#define INC_REGEN_RATE 0.002      /* Fraction of shield regained per second. */
#define COLLISION_PENALTY .75     /* Fraction of shield lost per second
				     of contact with a polyroid. */

#define NUM_KINDS 4               /* Number of colors of polyroids & lasers. */
#define NUM_HIGH_SCORES 14        /* Number of high scores to remember. */

#define POW_NONE 0
#define POW_LASER 1
#define POW_CONCUSSION 2
#define POW_REGEN 3
#define POW_NEVER 4


/*** Data structures. ********************************************************/

/* This is the structure which represents the spaceship. */
typedef struct {
  float centerX, centerY;     /* Midpoint of the triangle's spine. */
  float angle;                /* 0 <= angle < 2*PI. */
  float dx, dy, dTheta;       /* Current velocity and turn rate. */
  float ddx, ddy;             /* Current acceleration. */
  int score;                  /* Number of dead polyroids. */
  int concussions;            /* Number of concussions left. */
  float shields;              /* 0 <= shields <= 1. */
  short vertices[3][2];       /* The vertices of the triangle. */
  short thrustVertices[3][2]; /* The vertices of the thrust. */
  float shotStrength;
  float shotThrust;
  float regenRate;
} Ship;


/* The structure of a polyroid. */
typedef struct {
  int kind;                   /* What color. */
  float radius;               /* Radius. */
  float centerX, centerY;     /* Center coords. */
  float angle;                /* Current rotation angle. */
  float dx, dy, dTheta;       /* Positional and rotational velocities. */
  float ddx, ddy, ddTheta;    /* Positional and rotational accelerations. */
  int numSides;               /* Number of sides the polyroid has. */
  int dead;                   /* TRUE if the polyroid is dead. */
  int beingHit;               /* TRUE if the polyroid is being hit. */
  float chaseFactor;          /* From 0-1.  0 = Orbit, 1 = Chase. */
  int powerup;                /* Does the polyroid yield a special power? */
  short vertices[MAX_POLYROID_SIDES][2];  /* Vertices. */
} Polyroid;


/* A structure for holding a segment, including the slope. */
typedef struct {
  short *pt1;       /* Points to two shorts next to each other, x and y. */
  short *pt2;       /* Ditto. */
  float slope;      /* Slope of the segment. */
} Segment;


/* An element of the high score list. */
typedef struct hsNode {
  char *name;
  int score;
  struct hsNode *next, *prev;
} HSNode;


/*** Function declarations. **************************************************/

/* Configure: Sets up the window and graphics modes. */
void Configure(void);

/* Initializes drawing data. */
void DrawInit(void);

/* ClearWindow: Paints the window black. */
void ClearWindow(void);

/* Draw: Draws the screen. */
void Draw(float deltaT);

/* SetMessage sets a message for display. */
void SetMessage(int which, float howLong, char *msg);

/* Returns TRUE if segments intersect.  See util.c. */
int SegmentsIntersect(Segment *seg1, Segment *seg2);

/* Returns x, clipped to between the bounds. */
float bound(float x, float bottomBound, float topBound);

/* Calculates the new position of [polyroid, ship, laser]. */
void FindNewPolyroidPositions(float deltaT);
void FindNewShipPosition(float deltaT);
void FindNewLaserPosition(void);

/* Detects collisions, saving results in global variables. */
void CollisionDetect(float deltaT);

/* Returns the slope of the line containing the two points.  If the
   line is vertical, returns a large number. */
float CalculateSlope(short *pt1, short *pt2);

/* Allocates, initializes, and returns a new (random) Polyroid structure. */
Polyroid *MakeNewPolyroid(void);

/* These push and pop the laser stack out of and into currentLaser, so
that the user can push laser buttons at will. */
void ClearLaserStack(void);
void PushLaser(int whichLaser);
void PopLaser(int whichLaser);

/* Kills off a polyroid. */
void KillPolyroid(Polyroid *polyroid);

/* Returns the next 'viable' powerup, given a powerup. */
int NextPowerup(int current);

/* Ends a game. */
void EndGame(void);

/*** Global variable declarations. *******************************************/
/* All global variables are defined in polyroids.c. */

/* Which wave we're on. */
extern int wave;

/* Array of pointers to active polyroids. */
extern Polyroid *polyroids[MAX_POLYROIDS];

/* Number of active polyroids and number of living polyroids. */
extern int numPolyroids, numLivePolyroids;

/* Number of polyroids killed on this frame. */
extern int numNewPolyroidsNeeded;

/* Number of polyroids left to appear before end of wave. */
extern int numWavePolyroids;

/* TRUE if we need to create a polyroid with a powerup. */
extern Boolean needNewPowerup;

/* The player's ship. */
extern Ship ship;

/* TRUE if we need a redraw, FALSE otherwise. */
extern int needRedraw;

/* These booleans give the ship's control status. */
extern int thrustOn;
extern int turningLeft;
extern int turningRight;
extern int laserOn[4];
extern int concussionOn;

/* Which laser is active, -1 if none. */
extern int currentLaser;

/* Far endpoint of the laser. */
extern short laserEnd[2];

/* TRUE if the player is currently in contact with a polyroid. */
extern int dead;

/* TRUE if we are in attract mode. */
extern int attractMode;

/* In attract mode, TRUE for help screen, FALSE for high scores. */
extern int showHelp;

/* Pointers to the high score list. */
extern HSNode *firstScore;
extern HSNode *lastScore;
