/*

  position.c

  Position-calculation routines for polyroids.


    Copyright (C) 1993 Rob Shillingsburg
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


*/

#include "polyroids.h"
#include <sys/time.h>


/*** Interface function definitions: *****************************************/


void FindNewPolyroidPositions(float deltaT)
{
  int i, side;
  float newVelocity, allowedVelocity;
  float newDx, newDy;
  float angle;

  if (numWavePolyroids == 0 &&
      numNewPolyroidsNeeded == 0 && numPolyroids == 0) {
    char buf[1024];

    wave++;
    ship.shields = 1.0;
    numWavePolyroids = wave*wave - wave;
    InitPolyroids(wave);
    needNewPowerup = TRUE;
    sprintf(buf, "LEVEL %d", wave);
    SetMessage(0, 3.0, buf);
  }

  /* Make new polyroids, if we need them. */
  if (numNewPolyroidsNeeded) {
    static int prevPowerup = POW_LASER;

    for (i = 0;
	 numPolyroids < MAX_POLYROIDS && i < numNewPolyroidsNeeded;
	 i++) {
      polyroids[numPolyroids++] = MakeNewPolyroid();
      numLivePolyroids++;
      if (((rand() & 0xffff) < (0xffff / wave / 3))) {
	prevPowerup = NextPowerup(prevPowerup);
	polyroids[numPolyroids - 1]->powerup = prevPowerup;
	needNewPowerup = FALSE;
      }
    }
    numNewPolyroidsNeeded = 0;
  }

  for (i = 0; i < numPolyroids; i++) {
    Polyroid *polyroid = polyroids[i];

    /* Shrink polyroid if dead. */
    if (polyroid->dead) {
      polyroid->radius -= deltaT * SHRINK_SPEED;
      if (polyroid->radius < POLYROID_MIN_RADIUS) {
	if (numWavePolyroids > 0) {
	  numNewPolyroidsNeeded++;
	  numWavePolyroids--;
	}
	free(polyroids[i]);
	polyroids[i] = polyroids[numPolyroids-1];
	numPolyroids--;
	i--;
	continue;
      }
    }

    /* Find angle along which we'll accelerate. */
    if (attractMode) {  /* Use different accels for attract mode. */
      angle = atan2(ship.centerY - polyroid->centerY,
		    ship.centerX - polyroid->centerX);
    } else {
      /*angle = atan2((ship.centerY - polyroid->centerY)
		    * polyroid->chaseFactor - polyroid->dy,
		    (ship.centerX - polyroid->centerX)
		    * polyroid->chaseFactor - polyroid->dx);*/
      /* Try: d = distToShip, v = distance per second
	 chaseFactor = (d >= v) ? 1 : d/v; */

      angle = atan2((ship.centerY - (polyroid->centerY +
				     polyroid->dy * polyroid->chaseFactor)),
		    (ship.centerX - (polyroid->centerX +
				     polyroid->dx * polyroid->chaseFactor)));
    }

    /* Find acceleration. */
    polyroid->ddx =
      polyroid->numSides * PER_SIDE_POLYROID_ACCELERATION * cos(angle);
    polyroid->ddy =
      polyroid->numSides * PER_SIDE_POLYROID_ACCELERATION * sin(angle);

    /* Find component velocities. */
    newDx = polyroid->dx + polyroid->ddx * deltaT;
    newDy = polyroid->dy + polyroid->ddy * deltaT;
    
    /* Find vectored velocity and speed limit. */
    newVelocity = sqrt(newDy*newDy + newDx*newDx);
    allowedVelocity = bound(newVelocity,
			    polyroid->numSides * -PER_SIDE_POLYROID_VELOCITY,
			    polyroid->numSides * PER_SIDE_POLYROID_VELOCITY);
    
    /* If not exceeding speed limit, commit to new velocity,
       with drag adjustment. */
    /* If exceeding speed limit, use old velocity. */
    if (allowedVelocity == newVelocity) {
      polyroid->dx = newDx;
      polyroid->dy = newDy;
    }
    
    /* Find position, checking for edge conditions and enforcing speed
       limit on bounces.  Also advance the powerup. */
    polyroid->centerX += polyroid->dx * deltaT;
    if (bound(polyroid->centerX, 0, XSIZE) != polyroid->centerX) {
      polyroid->centerX = bound(polyroid->centerX, 0, XSIZE);
      polyroid->dx = -newDx * allowedVelocity / newVelocity;
      polyroid->dy = newDy * allowedVelocity / newVelocity;
      if (polyroid->powerup != POW_NEVER)
	polyroid->powerup = NextPowerup(polyroid->powerup);
    }
    polyroid->centerY += polyroid->dy * deltaT;
    if (bound(polyroid->centerY, 0, YSIZE) != polyroid->centerY) {
      polyroid->centerY = bound(polyroid->centerY, 0, YSIZE);
      polyroid->dx = newDx * allowedVelocity / newVelocity;
      polyroid->dy = -newDy * allowedVelocity / newVelocity;
      if (polyroid->powerup != POW_NEVER)
	polyroid->powerup = NextPowerup(polyroid->powerup);
    }

    /* Find new rotational velocity, checking against speed limit. */
    polyroid->dTheta += polyroid->ddTheta * deltaT;
    if (fabs(polyroid->dTheta) > MAX_POLYROID_SPIN && !polyroid->dead) {
      if (polyroid->numSides < MAX_POLYROID_SIDES) polyroid->numSides++;
      polyroid->dTheta = 0;
      polyroid->ddTheta = -polyroid->ddTheta;
    }

    /* Find new angle, mod 2*PI. */
    polyroid->angle += polyroid->dTheta * deltaT;
    if (polyroid->angle > 2*PI) polyroid->angle += -2*PI;
    if (polyroid->angle < -2*PI) polyroid->angle += 2*PI;
    
    /* Calculate vertices. */
    for (side = 0; side < polyroid->numSides; side++) {
      float baseAngle =  2*PI * side / polyroid->numSides;

      polyroid->vertices[side][X] = polyroid->centerX +
	polyroid->radius * cos(polyroid->angle + baseAngle);
      polyroid->vertices[side][Y] = polyroid->centerY +
	polyroid->radius * sin(polyroid->angle + baseAngle);
    }
  }
}


void FindNewShipPosition(float deltaT)
{
  /* Increase shields a little. */
  ship.shields =
    bound(ship.shields + deltaT * ship.regenRate, 0, 1);
  
  /* Find new angle, mod 2*PI. */
  ship.angle += ship.dTheta * deltaT;
  if (ship.angle >= 2*PI) ship.angle += -2*PI;
  if (ship.angle < 0) ship.angle += 2*PI;
  
  /* Find new acceleration, taking angle into account. */
  if (thrustOn) {
    ship.ddx = SHIP_THRUST * cos(ship.angle);
    ship.ddy = SHIP_THRUST * sin(ship.angle);
  }

  /* Find new velocities and positions, enforcing speed limit on bounces. */
  ship.dx += ship.ddx * deltaT - DRAG_COEFF * deltaT * ship.dx;
  ship.centerX += ship.dx * deltaT;
  if (ship.centerX != bound(ship.centerX, 0, XSIZE)) {
    ship.centerX = bound(ship.centerX, 0, XSIZE);
    ship.dx = bound(-ship.dx, -MAX_SHIP_VELOCITY, MAX_SHIP_VELOCITY);
  }
  ship.dy += ship.ddy * deltaT - DRAG_COEFF * deltaT * ship.dy;;
  ship.centerY += ship.dy * deltaT;
  if (ship.centerY != bound(ship.centerY, 0, YSIZE)) {
    ship.centerY = bound(ship.centerY, 0, YSIZE);
    ship.dy = bound(-ship.dy, -MAX_SHIP_VELOCITY, MAX_SHIP_VELOCITY);
  }
  
  /* Nose. */
  ship.vertices[0][X] = ship.centerX + SHIP_RADIUS * cos(ship.angle);
  ship.vertices[0][Y] = ship.centerY + SHIP_RADIUS * sin(ship.angle);

  /* Left corner. */
  ship.vertices[1][X] =
    ship.centerX + SHIP_RADIUS * cos(ship.angle + SHIP_CORNER_ANGLE);
  ship.vertices[1][Y] =
    ship.centerY + SHIP_RADIUS * sin(ship.angle + SHIP_CORNER_ANGLE);

  /* Right corner. */
  ship.vertices[2][X] =
    ship.centerX + SHIP_RADIUS * cos(ship.angle - SHIP_CORNER_ANGLE);
  ship.vertices[2][Y] =
    ship.centerY + SHIP_RADIUS * sin(ship.angle - SHIP_CORNER_ANGLE);

  /* Thrust tail. */
  ship.thrustVertices[0][X] = ship.centerX
    + SHIP_RADIUS * 2 * cos(ship.angle + PI);
  ship.thrustVertices[0][Y] = ship.centerY
    + SHIP_RADIUS * 2 * sin(ship.angle + PI);

  /* Thrust left corner. */
  ship.thrustVertices[1][X] =
    ship.centerX + SHIP_RADIUS * cos(ship.angle + SHIP_CORNER_ANGLE
				     + (PI - SHIP_CORNER_ANGLE) / 2);
  ship.thrustVertices[1][Y] =
    ship.centerY + SHIP_RADIUS * sin(ship.angle + SHIP_CORNER_ANGLE
				     + (PI - SHIP_CORNER_ANGLE) / 2);

  /* Thrust right corner. */
  ship.thrustVertices[2][X] =
    ship.centerX + SHIP_RADIUS * cos(ship.angle - SHIP_CORNER_ANGLE
				     - (PI - SHIP_CORNER_ANGLE) / 2);
  ship.thrustVertices[2][Y] =
    ship.centerY + SHIP_RADIUS * sin(ship.angle - SHIP_CORNER_ANGLE
				     - (PI - SHIP_CORNER_ANGLE) / 2);
}


void FindNewLaserPosition(void)
{
  if (currentLaser != -1) {
    float tanAngle = tan(ship.angle);  /* Cache this value. */

    if (tanAngle == 0.0) {  /* Straight across. */
      laserEnd[X] = (ship.angle == 0.0) ? XSIZE : 0;
      laserEnd[Y] = ship.centerY;
    } else {  /* Find intercept of laser with whichever wall. */
      float distToTop = (ship.angle < PI) ?
	YSIZE - ship.centerY : -ship.centerY;
      float topIntersectX = ship.centerX + distToTop / tanAngle;
      
      if (topIntersectX <= 0) {
	laserEnd[X] = 0;
	laserEnd[Y] = ship.centerY - ship.centerX * tanAngle;
      } else if (topIntersectX > XSIZE) {
	laserEnd[X] = XSIZE;
	laserEnd[Y] = ship.centerY + (XSIZE - ship.centerX) * tanAngle;
      } else {
	laserEnd[X] = topIntersectX;
	laserEnd[Y] = (ship.angle < PI) ? YSIZE : 0;
      }        
    }
  }
}
